#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_catalog.testing module

Package testing utilities.
"""

import sys

from persistent import Persistent
from zope.container.contained import Contained
from zope.interface import Attribute, Interface, implementer
from zope.schema import Datetime, Text, TextLine
from zope.schema.fieldproperty import FieldProperty

from pyams_i18n.schema import I18nTextLineField

__docformat__ = 'restructuredtext'


if sys.argv[-1].endswith('/bin/test'):

    class IContentInterface(Interface):
        """Content interface"""
        value = TextLine(title="Value property")
        first_date = Datetime(title="First date")
        keywords = Attribute("Keywords")
        facets = Attribute("Item facets values")
        text = Text(title="Text")
        empty = Text(title="Empty text")

    @implementer(IContentInterface)
    class MyContent(Persistent, Contained):
        """Content persistent class"""
        value = FieldProperty(IContentInterface['value'])
        first_date = FieldProperty(IContentInterface['first_date'])
        text = FieldProperty(IContentInterface['text'])
        empty = FieldProperty(IContentInterface['empty'])

        @property
        def keywords(self):
            """Keywords getter"""
            return ['category1', 'category2']

        @property
        def facets(self):
            """Facets getter"""
            return ['category:Content', 'price:0-100']

    class MyOtherContent(Persistent, Contained):
        """Other content class"""
        value = 'Other content value'

    class II18nContentInterface(Interface):
        """I18n content interface"""
        i18n_value = I18nTextLineField(title="I18n value property")

    @implementer(II18nContentInterface)
    class I18nContent(Persistent, Contained):
        """I18n content class"""
        i18n_value = FieldProperty(II18nContentInterface['i18n_value'])



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_catalog.zmi main module

"""

from hypatia.interfaces import ICatalog
from zope.interface import Interface

from pyams_utils.adapter import adapter_config
from pyams_zmi.interfaces import IAdminLayer, IObjectLabel

__docformat__ = 'restructuredtext'

from pyams_catalog import _  # pylint: ungrouped-imports


@adapter_config(required=(ICatalog, IAdminLayer, Interface),
                provides=IObjectLabel)
def catalog_label(context, request, view):
    """Catalog label getter"""
    return request.localizer.translate(_("Catalog"))



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_catalog.index module

This module provides several Hypatia indexes which use a discriminator based on interface
support of indexes objects.

WARNING: objects which don't implement the requested interface are not referenced in the index, including
in the "not indexed" part.
"""

from datetime import date, datetime

from ZODB.broken import Broken
from hypatia.facet import FacetIndex
from hypatia.field import FieldIndex
from hypatia.keyword import KeywordIndex
from hypatia.text import TextIndex
from hypatia.text.lexicon import Lexicon
from hypatia.util import BaseIndexMixin
from persistent import Persistent

from pyams_catalog.interfaces import DATE_RESOLUTION, NO_RESOLUTION
from pyams_catalog.nltk import NltkFullTextProcessor

__docformat__ = 'restructuredtext'


_MARKER = object()


class InterfaceSupportIndexMixin(BaseIndexMixin):
    """Custom index mixin handling objects interfaces"""

    def __init__(self, interface):
        self.interface = interface

    def discriminate(self, obj, default):
        """See interface IIndexInjection"""
        if self.interface is not None:
            obj = self.interface(obj, None)
            if obj is None:
                return default

        if callable(self.discriminator):
            value = self.discriminator(obj, _MARKER)
        else:
            value = getattr(obj, self.discriminator, _MARKER)
            if callable(value):
                value = value(obj)

        if (value is None) or (value is _MARKER):
            return default

        if isinstance(value, Persistent):
            raise ValueError('Catalog cannot index persistent object {0!r}'.format(value))

        if isinstance(value, Broken):
            raise ValueError('Catalog cannot index broken object {0!r}'.format(value))

        return value
    
    def index_doc(self, docid, obj):
        """See interface IIndexInjection"""
        if self.interface is not None:
            obj = self.interface(obj, None)
            if obj is None:
                return None
        return super().index_doc(docid, obj)


class FieldIndexWithInterface(InterfaceSupportIndexMixin, FieldIndex):
    """Field index with interface support"""

    def __init__(self, interface, discriminator, family=None):
        InterfaceSupportIndexMixin.__init__(self, interface)
        FieldIndex.__init__(self, discriminator, family)


def get_resolution(value, resolution):
    """Set resolution of given date or datetime

        >>> from pyams_catalog.interfaces import *
        >>> from pyams_catalog.index import get_resolution
        >>> from datetime import date, datetime

        >>> get_resolution(None, YEAR_RESOLUTION) is None
        True
        >>> get_resolution('', DATE_RESOLUTION)
        ''

    Starting with dates:

        >>> today = date(2017, 7, 11)
        >>> get_resolution(today, YEAR_RESOLUTION)
        datetime.date(2017, 1, 1)
        >>> get_resolution(today, MONTH_RESOLUTION)
        datetime.date(2017, 7, 1)
        >>> get_resolution(today, DATE_RESOLUTION)
        datetime.date(2017, 7, 11)
        >>> get_resolution(today, NO_RESOLUTION) is None
        True

    Asking for a resolution higher than DATE with a date input only returns date:

        >>> get_resolution(today, MINUTE_RESOLUTION)
        datetime.date(2017, 7, 11)

    Same examples with datetimes:

        >>> now = datetime(2017, 7, 11, 13, 22, 10)
        >>> get_resolution(now, YEAR_RESOLUTION)
        datetime.datetime(2017, 1, 1, 0, 0)
        >>> get_resolution(now, MONTH_RESOLUTION)
        datetime.datetime(2017, 7, 1, 0, 0)
        >>> get_resolution(now, DATE_RESOLUTION)
        datetime.datetime(2017, 7, 11, 0, 0)
        >>> get_resolution(now, HOUR_RESOLUTION)
        datetime.datetime(2017, 7, 11, 13, 0)
        >>> get_resolution(now, MINUTE_RESOLUTION)
        datetime.datetime(2017, 7, 11, 13, 22)
        >>> get_resolution(now, SECOND_RESOLUTION)
        datetime.datetime(2017, 7, 11, 13, 22, 10)
        >>> get_resolution(now, NO_RESOLUTION) is None
        True
    """
    if not value:
        return value
    if resolution < NO_RESOLUTION:
        args = []
        if not isinstance(value, datetime):
            resolution = min(resolution, DATE_RESOLUTION)
        args.extend(value.timetuple()[:resolution+1])
        if isinstance(value, datetime):
            args.extend(([1] * (DATE_RESOLUTION - resolution) + [0] * 5)[:7-len(args)])
            args.append(value.tzinfo)
            value = datetime(*args)
        else:
            args.extend([1] * (DATE_RESOLUTION - resolution))
            value = date(*args)
        return value
    return None


class DatetimeIndexWithInterface(FieldIndexWithInterface):
    """Normalized datetime index with interface support"""

    def __init__(self, interface, discriminator, resolution=DATE_RESOLUTION, family=None):
        FieldIndexWithInterface.__init__(self, interface, discriminator, family)
        self.resolution = resolution

    def discriminate(self, obj, default):
        value = super().discriminate(obj, default)
        return get_resolution(value, self.resolution)


class KeywordIndexWithInterface(InterfaceSupportIndexMixin, KeywordIndex):
    """Keyword index with interface support"""

    def __init__(self, interface, discriminator, family=None):
        InterfaceSupportIndexMixin.__init__(self, interface)
        KeywordIndex.__init__(self, discriminator, family)


class FacetIndexWithInterface(InterfaceSupportIndexMixin, FacetIndex):
    """Facet index with interface support"""

    def __init__(self, interface, discriminator, facets, family=None):
        InterfaceSupportIndexMixin.__init__(self, interface)
        FacetIndex.__init__(self, discriminator, facets, family)


class TextIndexWithInterface(InterfaceSupportIndexMixin, TextIndex):
    """Text index with interface support"""

    # pylint: disable=too-many-arguments
    def __init__(self, interface, discriminator, lexicon=None, language='english',
                 index=None, family=None):
        InterfaceSupportIndexMixin.__init__(self, interface)
        if lexicon is None:
            lexicon = Lexicon(NltkFullTextProcessor(language))
        TextIndex.__init__(self, discriminator, lexicon, index, family)



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_catalog.query module
"""

from hypatia.query import Comparator, Query
from zope.intid.interfaces import IIntIds

from pyams_utils.registry import query_utility


__docformat__ = 'restructuredtext'


class IsNone(Comparator):
    """Index query search for None values"""

    def __init__(self, index):
        super().__init__(index, None)

    def _apply(self, name):
        return self.index.not_indexed()


class ResultSet:
    """Base catalog query result set wrapper

    This class just wraps a query result to return final objects
    instead of internal IDs.
    """

    def __init__(self, query):
        self.query = query
        self.intids = query_utility(IIntIds)

    def __iter__(self):
        intids = self.intids
        if intids is not None:
            query = self.query
            if isinstance(query, Query):
                query = query.execute()
            if isinstance(query, tuple):
                query = query[1]
            for oid in query:
                yield intids.queryObject(oid)


class CatalogResultSet:
    """Catalog query result set wrapper"""

    def __init__(self, query):
        self.query = query
        self.intids = query_utility(IIntIds)
        self.first = []
        self.last = []

    def __iter__(self):
        for item in self.first:
            yield item
        intids = self.intids
        if intids is not None:
            query = self.query
            if isinstance(query, Query):
                query = query.execute()
            if isinstance(query, tuple):
                query = query[1]
            for oid in query:
                if isinstance(oid, int):
                    target = intids.queryObject(oid)
                    if target is not None:
                        yield target
                else:
                    yield oid
        for item in self.last:
            yield item

    def __len__(self):
        if isinstance(self.query, (list, tuple, ResultSet)):
            nb_docs = len(self.query)
        else:
            nb_docs = self.query[0]
        return len(self.first) + nb_docs + len(self.last)

    def prepend(self, items):
        """Insert a list of elements at the beginning of the results set"""
        insert = self.first.insert
        for index, item in enumerate(items):
            insert(index, item)

    def append(self, items):
        """Append a list of elements at the end of the results set"""
        append = self.last.append
        for item in items:
            append(item)


def or_(source, added):
    """Combine two queries with 'or'"""
    if source is None:
        source = added
    else:
        source |= added
    return source


def and_(source, added):
    """Combine two queries with 'and'"""
    if source is None:
        source = added
    else:
        source &= added
    return source



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_catalog.nltk module

"""

import nltk
from hypatia.text import Lexicon
from hypatia.text.interfaces import ISplitter
from zope.interface import implementer

from pyams_i18n.language import BASE_LANGUAGES
from pyams_utils.unicode import translate_string


__docformat__ = 'restructuredtext'


@implementer(ISplitter)
class NltkStemmedTextProcessor:
    """NLTK based text processor using stemmer"""

    def __init__(self, language='english'):
        if language in BASE_LANGUAGES:
            language = BASE_LANGUAGES[language].lower()
        self.language = language
        self.stemmer = nltk.stem.SnowballStemmer(language, ignore_stopwords=True)

    def process(self, lst):
        """Main process method"""
        result = []
        for s in lst:  # pylint: disable=invalid-name
            translated = translate_string(s, keep_chars="'-").replace("'", ' ')
            tokens = nltk.word_tokenize(translated, self.language)
            result += [stem for stem in [self.stemmer.stem(token) for token in tokens
                                         if token not in self.stemmer.stopwords]
                       if stem and (len(stem) > 1) and (stem not in self.stemmer.stopwords)]
        return result

    def processGlob(self, lst):  # pylint: disable=invalid-name
        """Globs processing method"""
        result = []
        for s in lst:  # pylint: disable=invalid-name
            translated = translate_string(s, keep_chars="'-*?").replace("'", ' ')
            tokens = nltk.word_tokenize(translated, self.language)
            result += [stem for stem in [self.stemmer.stem(token) for token in tokens
                                         if token not in self.stemmer.stopwords]
                       if stem and (len(stem) > 1) and (stem not in self.stemmer.stopwords)]
        return result


@implementer(ISplitter)
class NltkFullTextProcessor:
    """NLTK based full text processor"""

    def __init__(self, language='english'):
        if language in BASE_LANGUAGES:
            language = BASE_LANGUAGES[language].lower()
        self.language = language

    def process(self, lst):
        """Main processing method"""
        result = []
        for s in lst:  # pylint: disable=invalid-name
            translated = translate_string(s, keep_chars="'-").replace("'", ' ')
            result += [
                '{}*'.format(token)
                for token in nltk.word_tokenize(translated, self.language)
                if token and len(token) > 1
            ]
        return result

    def processGlob(self, lst):  # pylint: disable=invalid-name
        """Globs processing method"""
        result = []
        for s in lst:  # pylint: disable=invalid-name
            translated = translate_string(s, keep_chars="'-*?").replace("'", ' ')
            result += [
                '{}*'.format(token)
                for token in nltk.word_tokenize(translated, self.language)
                if token and len(token) > 1
            ]
        return result


def get_fulltext_lexicon(language):
    """Fulltext lexicon creator

    :param language: lexicon language code, defined with two lowercase letters ('en', 'fr'...)
    """
    return Lexicon(NltkFullTextProcessor(language=language))
